;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                 LCD-Routinen                ;;
;;                 ============                ;;
;;              (c)andreas-s@web.de            ;;
;;                                             ;;
;; 4bit-Interface                              ;;
;; DB4-DB7:       PC0-PC3                      ;;
;; RS:            PC4                          ;;
;; E:             PC5                          ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;
;sendet ein Datenbyte an das LCD
lcd_data:
	push	temp1
	push	temp2
	push	temp3
	in 		srsk,sreg
	push	srsk

	mov		temp2,temp1				;"Sicherungskopie" fr die bertragung des 2.Nibbles
	swap	temp1					;Vertauschen
	andi	temp1,0b00001111		;oberes Nibble auf Null setzen
	sbr		temp1,1<<PIN_RS			;entspricht 0b00010000
	in		temp3,LCD_PORT
	andi	temp3,0x80
	or		temp1,temp3
	out		LCD_PORT,temp1			;ausgeben
	call	lcd_enable				;Enable-Routine aufrufen
									;2.Nibble, kein swap da es schon an der richtigen stelle ist
	andi	temp2,0b00001111		;obere Hlfte auf Null setzen 
	sbr		temp2,1<<PIN_RS			;entspricht 0b00010000
	or		temp2,temp3
	out		LCD_PORT,temp2			;ausgeben
	call	lcd_enable				;Enable-Routine aufrufen
	call	delay50us				;Delay-Routine aufrufen

	pop		srsk
	out		sreg,srsk
	pop		temp3
	pop		temp2
	pop		temp1
	ret								;zurck zum Hauptprogramm

;sendet einen Befehl an das LCD
lcd_command:						;wie lcd_data, nur ohne RS zu setzen
	push	temp1
	push	temp2
	push	temp3
	in 		srsk,sreg
	push	srsk

	mov		temp2,temp1
	swap	temp1
	andi	temp1,0b00001111
	in		temp3,LCD_PORT
	andi	temp3,0x80
	or		temp1,temp3
	out		LCD_PORT,temp1
	call	lcd_enable
	andi	temp2,0b00001111
	or		temp2,temp3
	out		LCD_PORT,temp2
	call	lcd_enable
	call	delay50us
 
	pop		srsk
	out		sreg,srsk
	pop		temp3
	pop		temp2
	pop		temp1
	ret

;erzeugt den Enable-Puls
lcd_enable:
	sbi		LCD_PORT,PIN_E			;Enable high
	nop								;3 Taktzyklen warten
	nop
	nop
	cbi		LCD_PORT,PIN_E			;Enable wieder low
	ret								;Und wieder zurck                     

;Pause nach jeder bertragung
delay50us:							;50us Pause
	push	temp1
	in		srsk,sreg
	push	srsk

	ldi		temp1,(XTAL*41/5)/1000000
	nop
delay50us_:
	dec		temp1
	nop
	nop
	nop
	brne	delay50us_

	pop		srsk
	out		sreg,srsk
	pop		temp1
	ret								;wieder zurck
 
;Lngere Pause fr manche Befehle
delay5ms:							;5ms Pause
	push	temp1
	push	temp2
	in		srsk,sreg
	push	srsk
		
	ldi		temp2,$C9
WGLOOP0:	
	ldi		temp1,(XTAL*5/606)/1000
WGLOOP1:	
	dec		temp1
	brne	WGLOOP1
	dec		temp2
	brne	WGLOOP0

	pop		srsk
	out		sreg,srsk
	pop		temp2
	pop		temp1
	ret								;wieder zurck
 
;noch Lngere Pause fr manche Befehle
delay100ms:
	push	temp1
	push	temp2
	push	temp3							;100ms Pause
	in		srsk,sreg
	push	srsk

	ldi	temp3,$04
WGLOOP2:	
	ldi	temp2,$92
WGLOOP3:
	ldi	temp1,(XTAL*6/42)/10000
WGLOOP4:	
	dec	temp1
	brne	WGLOOP4
	dec		temp2
	brne	WGLOOP3
	dec		temp3
	brne	WGLOOP2

	pop		srsk
	out		sreg,srsk
	pop		temp3
	pop		temp2
	pop		temp1
	ret								;wieder zurck

;Initialisierung: muss ganz am Anfang des Programms aufgerufen werden
lcd_init:
	push	temp1
	push	temp3
	in		srsk,sreg
	push	srsk

	in		temp1,LCD_DDR
	ori		temp1,(1<<PIN_E)|(1<<PIN_RS)|0x0F
	out		LCD_DDR,temp1
	ldi		temp3,6
powerupwait:
	call	delay5ms
	call	delay5ms
	dec		temp3
	brne	powerupwait
	ldi		temp1,0b00000011		;muss 3mal hintereinander gesendet
	out		LCD_PORT,temp1			;werden zur Initialisierung
	call	lcd_enable				;1
	call	delay5ms
	call	delay5ms
	call	lcd_enable				;2
	call	delay5ms
	call	delay5ms
	call	lcd_enable				;und 3!
	call 	delay5ms
	call	delay5ms
	ldi		temp1,0b00000010		;4bit-Modus einstellen
	out		LCD_PORT,temp1
	call	lcd_enable
	call	delay5ms
	call	delay5ms
	ldi		temp1,0b00101000		;4 Bot, 2 Zeilen
	call	lcd_command
	ldi		temp1,0b00001100		;Display on, Cursor off
	call 	lcd_command
	ldi   	temp1,0b00000100		;endlich fertig
	call 	lcd_command

	pop		srsk
	out		sreg,srsk
	pop		temp3
	pop		temp1
	ret
 
;Sendet den Befehl zur Lschung des Displays
lcd_clear:
	push	temp1

	ldi		temp1,0b00000001		;Display lschen
	call	lcd_command
	call	delay5ms
	call	lcd_home

	pop		temp1
	ret

;Cursor Home
lcd_home:
	push	temp1

	ldi		temp1,0b00000010		;Cursor Home
	call	lcd_command
	call	delay5ms

	pop		temp1
	ret

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

; Ausgeben eines Strings aus dem SRAM ans LCD-Display
Call_SRAM_String_to_LCD:
	push	temp1
	push	temp2
	in		srsk,sreg
	push	srsk

Call_SRAM_String_to_LCD_0:
	ld		temp1,Y+					;ldt Zeichen aus Sram (von Position des Y-Zeigers)
	cpi		temp2,0						;ist dieses Zeichen 0 ???
	breq	Call_SRAM_String_to_LCD_1	;wenn ja springe zum angegebenen Label
	call	lcd_data					;wenn nicht mache hier normal weiter - gib das Zeichen an LCD
	dec		temp2	
	rjmp	Call_SRAM_String_to_LCD_0	;Springe an den Anfang der Schleife fr nchstes Zeichen
Call_SRAM_String_to_LCD_1:				;Label fr ende der Schleife - wird angesprungen, wenn Zeichen 0 ist

	pop		srsk
	out		sreg,srsk
	pop		temp2
	pop		temp1
	ret									;Rcksprung zur Herkunftsstelle

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

;Einen konstanten Text aus dem Flash Speicher ausgeben. Der Text wird mit einer 0 beendet
lcd_flash_string:
	push	temp1						;sichere Temp1 auf den Stack
	push	temp2						;sichere Temp2 auf den Stack
	in 		srsk,sreg
	push	srsk

lcd_flash_string_1:   					;Temp2 wird als Zeichenzhler verwendet und kann schon ausserhalb gesetzt werden
	cpi		temp2,0						;ist Zhler = 0 
	breq	lcd_zeile1					;springe zu Aufruf fr Zeile1 (Cursorposition)
	cpi		temp2,16					;ist Zhler = 16 
	breq	lcd_zeile2					;springe zu Aufruf fr Zeile2 (Cursorposition)
	cpi		temp2,32					;ist Zhler = 32
	breq	lcd_zeile3					;springe zu Aufruf fr Zeile3 (Cursorposition)
	cpi		temp2,48					;ist Zhler = 48
	breq	lcd_zeile4					;springe zu Aufruf fr Zeile4 (Cursorposition)
lcd_flash_string_1a:
	lpm		temp1,Z+					;ruft String aus Flash ab (von Position des Z-Zeiger)
	inc		temp2						;zhlt Zeichenzhler hoch
	cpi		temp1,0						;ist zeichen 0? 
	breq	lcd_flash_string_2			;ja, dann gehe zu diesem Label
	call	lcd_data					;wenn nicht mache hier weiter und gib Zeichen an das LCD
	jmp		lcd_flash_string_1			;Rcksprung zum anfang der Schleife fr nchstes Zeichen
lcd_flash_string_2:						;Label fr Ende-Ansprung

	pop		srsk
	out		sreg,srsk
	pop		temp2						;lade Temp2 vom Stack
	pop		temp1						;lade Temp1 vom Stack
	ret									;Rcksprung zur Herkunftsstelle

; Unterprogramme zum Zeilensprung bei durchgehendem Text
lcd_zeile1:
	push	temp1						;sichere auf Stack
	ldi		temp1,zeile1-2				;Zeilenanfang Zeile 1
	call	lcd_command					;Cursor setzen
	pop		temp1						;hole vom Stack
	jmp		lcd_flash_string_1a			;zurck zur Zeichenausgabe
lcd_zeile2:
	push	temp1
	ldi		temp1,zeile2-2
	call	lcd_command
	pop		temp1
	jmp		lcd_flash_string_1a
lcd_zeile3:
	push	temp1
	ldi		temp1,zeile3-2
	call	lcd_command
	pop		temp1
	jmp		lcd_flash_string_1a
lcd_zeile4:
	push	temp1
	ldi		temp1,zeile4-2
	call	lcd_command
	pop		temp1
	jmp		lcd_flash_string_1a


;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


;Eine 16-Bit-Zahl aus dem Register temp1:temp2 mit Vorzeichen dezimal ausgeben
;bentigt 7 Zeichen (inkl. Vorzeichen und 1 Festkommastelle)
lcd_number:
	push	temp1							;sichere auf Stack
	push	temp2							;sichere auf Stack
	push	temp3							;sichere auf Stack
	push	temp4							;sichere auf Stack
	push	wL								;sichere auf Stack
	push	wH								;sichere auf Stack
	in		srsk,SREG						;sichere SREG in Hilfsregister
	push	srsk							;sichere Hilfsregister auf Stack

	mov		temp3,temp1						;setze Zahl in Temp3(L) und Temp4(H)
	mov		temp4,temp2

	ldi		temp1,zeile1
	ldi		temp2,9
	add		temp1,temp2	;Displayausgabe festlegen
	call	lcd_command

			;hier beginnt ziffernausgabe

					;Vorzeichen setzen

	ldi		temp1,'0'
	sbrs	temp4,7							;wenn Bit 7=1(neg) berspringe nchste zeile
	jmp		lcd_number_wert_a

	ldi		temp1,'-'						;setze minuszeichen
	call	lcd_data
	neg		temp3
	neg		temp4							;negiere register um wert zu erhalten
	subi	temp4,1
	jmp		lcd_number_wert_b				;berspringe Lehrzeichen, weil Wert negativ
lcd_number_wert_a:
	ldi		temp1,' '						;setze Leerzeichen
	call	lcd_data
lcd_number_wert_b:

					;Zehntausender berechnen  
							      		  
	ldi 	wL,LOW(10000)
	ldi 	wH,HIGH(10000)
	ldi		temp1,'0'				;abzhlen wieviele Zehntausender in der Zahl enthalten sind
lcd_number_1:
    sub		temp3,wL				;10000 abziehen um Anzahl zu bestimmen
	sbc		temp4,wH					
	brcs	lcd_number_1a			;kleiner als null (Carry gesetzt?) -> gehe zur Ausgabe
	inc		temp1
	jmp		lcd_number_1                       
lcd_number_1a:						; die Zehntausenderstelle ausgeben
	cpi 	temp1,'0'               ; Compare r19 with 0
	brne 	lcd_number_1b
	ldi		temp1,' '
lcd_number_1b:
	mov		temp2,temp1                                           
;	call	lcd_data				;hier einschalten, wenn 10000er ausgegeben werden sollen
	add		temp3,wL	            ; 10000 wieder dazuzhlen, da die
    adc		temp4,wH             ; vorherhgehende Schleife 10000 zuviel abgezogen hat

					;Tausender berechnen  
							      		  
	ldi 	wL,LOW(1000)
	ldi 	wH,HIGH(1000)
	ldi		temp1,'0'				;abzhlen wieviele Zehntausender in der Zahl enthalten sind
lcd_number_2:
    sub		temp3,wL				;1000 abziehen um Anzahl zu bestimmen
	sbc		temp4,wH					
	brcs	lcd_number_2a			;kleiner als null (Carry gesetzt?) -> gehe zur Ausgabe
	inc		temp1
	jmp		lcd_number_2                               
lcd_number_2a:						; die Tausenderstelle ausgeben
	cpi 	temp1,'0'
	brne	lcd_number_2b
	cpi		temp2,' '
	brne	lcd_number_2b
	ldi		temp1,' '	            ; Compare r19 with 3                                
lcd_number_2b:
	mov		temp2,temp1
;	call	lcd_data				;hier einschalten, wenn 1000er Stelle ausgegeben werden soll
	add		temp3,wL	            ; 1000 wieder dazuzhlen, da die
    adc		temp4,wH             ; vorhergehende Schleife 1000 zuviel abgezogen hat
                                  		  
					;Hunderter berechnen  
							      		                                            
	ldi 	wL,LOW(100)
	ldi 	wH,HIGH(100)
	ldi		temp1,'0'				; abzhlen wieviele Hunderter in der Zahl enthalten sind
lcd_number_3:
    sub		temp3,wL				;100 abziehen um Anzahl zu bestimmen
	sbc		temp4,wH					
	brcs	lcd_number_3a
	inc		temp1
	jmp		lcd_number_3                                 
lcd_number_3a:						; die Hunderterstelle ausgeben
	cpi 	temp1,'0'               ; Compare r19 with 3
	brne 	lcd_number_3b
	cpi		temp2,' '
	brne	lcd_number_3b
	ldi		temp1,' '                                           
lcd_number_3b:
	call	lcd_data
	add		temp3,wL	            ; 100 wieder dazuzhlen, da die
    adc		temp4,wH             ; vorhergehende Schleife 100 zuviel abgezogen hat
                                          
					;Zehner berechnen  
							      		                                                                                     
	ldi		temp1,'0'				; abzhlen wieviele Zehner in der Zahl enthalten sind
lcd_number_4:
	subi	temp3,10
	brcs	lcd_number_4a
	inc		temp1
	jmp		lcd_number_4                                   
lcd_number_4a:						; die Zehnerstelle ausgeben
	call	lcd_data
	subi	temp3,-10               ; 10 wieder dazuzhlen, da die vorhergehende Schleife 10 zuviel abgezogen hat

					;Kommastelle setzen  
							      		        
	ldi		temp1,','				;Kommastelle setzen
	call	lcd_data

					;die brig gebliebenen Einer noch ausgeben

	ldi		temp1,'0'						
	add		temp1,temp3
	call	lcd_data

	pop		srsk
	out		SREG,srsk
	pop		wH
	pop		wL
	pop 	temp4
	pop		temp3
	pop		temp2
	pop		temp1
	ret


;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


;eine Zahl aus dem Register temp1 hexadezimal ausgeben
lcd_number_hex:
	push	temp1
	
	swap	temp1
	andi	temp1,$0F
	call	lcd_number_hex_digit

	pop		temp1
	push	temp1
	
	andi	temp1,$0F
	call	lcd_number_hex_digit

	pop		temp1
	ret

lcd_number_hex_digit:
	cpi		temp1,10
	brlt	lcd_number_hex_digit_1
	subi	temp1,-('A'-'9'-1)
lcd_number_hex_digit_1:
	subi	temp1,-'0'
	call	lcd_data
	ret



;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

lcd_Uhr_Ziffern:
        push    temp2               ; register sichern,
                                    ; wird fr Zwsichenergebnisse gebraucht     
        ldi     temp2, '0'         
lcd_Uhr_Ziffern_10:                
        subi    temp1, 10           ; abzhlen wieviele Zehner in
        brcs    lcd_Uhr_Ziffern_1        ; der Zahl enthalten sind
        inc     temp2
        jmp     lcd_Uhr_Ziffern_10
lcd_Uhr_Ziffern_1:
        push    temp1               ; den Rest sichern (http://www.mikrocontroller.net/topic/172026)
        mov     temp1,temp2         ; 
        call    lcd_data            ; die Zehnerstelle ausgeben
        pop     temp1               ; den Rest wiederherstellen
        subi    temp1, -10          ; 10 wieder dazuzhlen, da die
                                    ; vorhergehende Schleife 10 zuviel
                                    ; abgezogen hat
                                    ; das Subtrahieren von -10
                                    ; = Addition von +10 ist ein Trick
                                    ; da kein addi Befehl existiert
        ldi     temp2, '0'          ; die brig gebliebenen Einer
        add     temp1, temp2        ; noch ausgeben
        call   lcd_data
 
        pop     temp2               ; Register wieder herstellen
        ret
